﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Data.OleDb;
using System.IO;
using System.Windows.Forms;

namespace Interpolace
{
    //TODO: Program nebyl nikdy radne otestovan. Pouziti jen na vlastni nebezpeci.
    /// <summary>
    /// Main class for start of the program
    /// </summary>
    class Program
    {
        /// <summary>
        /// Startovaci metoda. Bud probehne program v prikazovem radku nebo se zobrazi GUI.
        /// </summary>
        static void Main(string[] args)
        {
            if (args.Count() == 0)
            {
                //Spusti se dialogove okno pro ovladani
                Prostredi.ui = "Windows"; //uzivatel pracuje z GUI
                Application.EnableVisualStyles();
                Application.SetCompatibleTextRenderingDefault(false);
                Application.Run(new Hlavni());
            }
            if (args.Count() == 1)
            {
                //Vypise se informace o pouziti programu z prikazove radky
                Console.WriteLine("Pouziti: Interpolace.exe -w nazev_souboru_bez_pripony [pocet_bodu]");
                Console.WriteLine("Prepinac -w zajistiti vygenerovani nahodnych bodu. Body zapise do textoveho, binarniho (3xdouble pro kazdy bod) a accdb souboru.");
                Console.WriteLine("Pokud neni uveden pocet bodu bude vygenerovano 50 bodu.");
                Console.WriteLine("Pouziti: Interpolace.exe -r nazev_souboru_ke_cteni x y [exponent] [pocet_bodu_v_okoli]");
                Console.WriteLine("Prepinac -r nacte data ze souboru a provede interpolaci IDW na zolene souradnici.");
                Console.WriteLine("Pokud neni uveden exponent je pouzita hodnota 2.");
                Console.WriteLine("Pokud neni uveden pocet bodu v okoli bere se 10 nejblizsich.");
            }
            if (args.Count() > 1)
            {
                Interpolace inter = new Interpolace();
                Prostredi.ui = "Console"; //Uzivatel pracuje z prikazove radky
                if (args[0].Equals("-w"))
                {
                    int pocet = 50;
                    try
                    {
                        if (args.Count() == 3) pocet = Int32.Parse(args[2]);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Chybne zadany pocet bodu. Bude vygenerovano 50 bodu. " + ex.Message);
                    }
                    //Budeme generovat body a ulozime je do souboru
                    //Budeme ukladat jak do textoveho tak do binarniho a i do databaze
                    inter.generovatNahodne(pocet);
                    inter.ulozitDoSouboru(args[1]);
                }
                if (args[0].Equals("-r"))
                {
                    double exponent = 2;
                    int okoli = 10;
                    double x = 0;
                    double y = 0;
                    try
                    {
                        if (args.Count() > 2) x = Double.Parse(args[2]);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Chybne zadane x. Bude pouzita hodnota 0. " + ex.Message);
                    }
                    try
                    {
                        if (args.Count() > 3) y = Double.Parse(args[3]);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Chybne zadane y. Bude pouzita hodnota 0. " + ex.Message);
                    }
                    try
                    {
                        if (args.Count() > 4) exponent = Double.Parse(args[4]);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Chybne zadany exponent. Bude pouzita hodnota 2. " + ex.Message);
                    }
                    try
                    {
                        if (args.Count() > 5) okoli = Int32.Parse(args[5]);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Chybne zadany pocet bodu v okoli. Bude pouzita hodnota 10. " + ex.Message);
                    }
                    //Budeme cist body ze souboru
                    //O jaky typ souboru se jedna budeme rozhodovat na zaklade pripony
                    //Muzeme pridat i kontrolu obsahu
                    if (inter.nacist(args[1])) {
                        //Pokud se nepodari nacist soubor interpolace se neprovede
                        Vysledek v = inter.interpolovat(x, y, exponent, okoli);
                        if (v.status.Equals("KO")) { 
                            //ikdyz se soubor nacetl muze pri interpolaci dojit k nejake chybe
                            Console.WriteLine("Interpolovana hodnota nemohla byt urcena. Nastala neocekavana chyba.");
                        } else {
                            //pokud je interpolace dokoncena zapiseme
                            Console.WriteLine("Interpolovana hodnota: " + v.hodnota);
                        }
                    }
                }
            }
        }
    }

    /// <summary>
    /// Trida pro zaznam informace o jednom mereni. Bylo by mozne vyuzit struct, ale protoze jsme se struct nezabyvali, tak je pouzita trida.
    /// </summary>
    class Mereni
    {
        //souradnicovy system se neresi, pouziva se bezny euklidovsky prostor
        public double X; //souradnice X
        public double Y; //souradnice Y
        public double H; //hodnota mereni v bode
    }
    /// <summary>
    /// Hlavni trida pro interpolaci.
    /// </summary>
    class Interpolace
    {
        ArrayList kolekceMereni = new ArrayList(); //Obsahuje seznam mereni

        /// <summary>
        /// Metoda generuje nahodne mereni s hodnotami souradnic X a Y a hodnot mereni z intervalu 0 - 1.
        /// </summary>
        public void generovatNahodne(int pocet) {
            Random randObj = new Random(); //Trida pro generovani pseudonahodnych cisel
            kolekceMereni.Clear(); //Z kolekce mereni jsou pred vygenerovanim odstraneny vsechna mereni. 
            for (int i = 0; i < pocet; i++) //Vygeneruje se presny pocet mereni
            {
                Mereni m = new Mereni(); 
                m.X = randObj.NextDouble(); //Nove nahodne cislo se vyuzije pro nastaveni souradnice X
                m.Y = randObj.NextDouble(); //Nove nahodne cislo se vyuzije pro nastaveni souradnice Y
                m.H = randObj.NextDouble(); //Nove nahodne cislo se vyuzije pro nastaveni hodnoty mereni
                kolekceMereni.Add(m); //Mereni se vlozi do kolekece
                if (Prostredi.ui.Equals("Console")) { //Pokud pracuje uzivatel v prikazovem radku, tak se mu zobrazuje zperna informace o stavu generovani
                    Console.WriteLine("Generuji " + pocet + " bodu");
                    Console.WriteLine("Bod " + i + ": " + m.X + " " + m.Y + " " + m.H);
                }
            }
        }

        /// <summary>
        /// Metoda ulozi kolekci mereni do tri souboru. Tak abychom demonstrovali zapis do textoveho a binarniho souboru a do databaze 
        /// </summary>
        /// <param name="filename">filename v tomto pripade predstavuje cast nazvu bez pripony, ktery je pouzit pro vsechny sohbory</param>
        public void ulozitDoSouboru(String filename) {
            //TODO: jeste je vhodne osetrit, kdyz uzivatel pojmenuje soubor Mereni, to by pak doslo k problemum v pripade databaze
            StreamWriter sw = null;
            BinaryWriter bw = null;
            OleDbConnection con = null;
            try {
                FileStream fs = new FileStream(filename + ".csv", FileMode.Create); //Ukazatel do textoveho souboru
                sw = new StreamWriter(fs);
                FileStream fs2 = new FileStream(filename + ".bin", FileMode.Create); //Ukazatel do binarniho souboru
                bw = new BinaryWriter(fs2);
                con = new OleDbConnection(); //Pripojeni do databaze
                if (File.Exists(filename + ".accdb")) {
                    File.Delete(filename + ".accdb"); //Databaze je nejdrive smazana, pokud jiz existuje
                }
                File.Copy("Mereni.accdb", filename + ".accdb"); //Ze vzorove databaze s prazdnou tabulkou Mereni je vytvorena kopie se kterou se bude pracovat
                con.ConnectionString = "Provider=Microsoft.ACE.OLEDB.12.0;Data Source=.\\" + filename + ".accdb";
                con.Open();
                //Prochazi se kolekce mereni
                foreach (Mereni m in kolekceMereni)
                {
                    sw.WriteLine(m.X + ";" + m.Y + ";" + m.H); //zapis radku do textoveho souboru
                    bw.Write(m.X); //binarni zapis
                    bw.Write(m.Y);
                    bw.Write(m.H);
                    //zapis do tabulky v MS Access
                    String SQL = "INSERT INTO Mereni (X, Y, H) VALUES (" + m.X + ", " + m.Y + ", " + m.H + ")";
                    OleDbCommand cmdinsert = new OleDbCommand(SQL, con);
                    cmdinsert.ExecuteNonQuery();
                } 
            //TODO: pokud nastane chyba u libovolneho typu vystupu, pak se zapis ukonci ve vsech vystupech. Toto by bylo mozno snadno osetrit rodelenim zapisu do vice metod, aplikace by se tim mozna trochu zpomalila.
            } catch(IOException iex) {
                //Nastane li chyba pri zapisu do souboru, pak je o tom uzivatel informovan
                Prostredi.ShowMessage("Neni mozne zapsat data do souboru. " + iex.Message);
            } catch (OleDbException oex){
                //Nastane li chyba pri zapisu do databaze, pak je o tom uzivatel informovan
                Prostredi.ShowMessage("Neni mozne zapsat data do databaze. " + oex.Message);
            }
            //Pokud nastane chyba, je vhodne uzavrit otevrene soubory a spojeni. 
            //TODO: i pri uzavirani muze nastat chyba, takze by i tyto tri radky mely byt v bloku try
            finally {
                sw.Close();
                bw.Close();
                con.Close();
            }
        }

        /// <summary>
        /// Metoda rozhoduje dle pripony souboru o tom jaky typ souboru se jedna. Na zaklade toho spusti prislusnou metodu pro nacteni.
        /// </summary>
        /// TODO: Rozhodovani na zaklade pripony je nevhodne, ale rychle.
        public bool nacist(String path) {
            String end = path.Substring(path.Length -3, 3);
            switch (end) {
                case "csv":
                    nacistZTextovehoSouboru(path);
                    break;
                case "txt":
                    nacistZTextovehoSouboru(path);
                    break;
                case "bin":
                    nacistZBinarnihoSouboru(path);
                    break;
                case "cdb":
                    nacistZDatabaze(path);
                    break;
                default:
                    Prostredi.ShowMessage("Oekavam soubor s priponou txt, csv, bin nebo accdb");
                    return false;
            }
            //Kontrola zda se nejaka mereni podarilo nacist
            if (kolekceMereni.Count == 0)
            {
                Prostredi.ShowMessage("Soubor zrejme neobsahuje zadna mereni");
                return false;
            }
            else {
                return true;
            }
        }

        /// <summary>
        /// Metoda nacita mereni z textoveho souboru.
        /// </summary>
        public void nacistZTextovehoSouboru(String path) {
            kolekceMereni.Clear();
            Console.WriteLine("Text");
            if (File.Exists(path))
            {
                //Nacti
                StreamReader sr = null;
                try
                {
                    FileStream fs = new FileStream(path, FileMode.Open);
                    sr = new StreamReader(fs);
                    String line;
                    //Soubor se cte radek po radku
                    while ((line = sr.ReadLine()) != null)
                    {
                        //Nacteny radek se rozdeli na zaklade znaku, kterym je strednik
                        //TODO: vhodne doplnit o jiny libovolny znak dle volby uzivatele
                        string[] split = line.Split(new Char[] { ';' });
                        Mereni m = new Mereni();
                        m.X = Double.Parse(split[0]);
                        m.Y = Double.Parse(split[1]);
                        m.H = Double.Parse(split[2]);
                        kolekceMereni.Add(m); //Nactene mereni se prida do kolekce
                    }
                }
                catch (IOException iex)
                {
                    Prostredi.ShowMessage("Neni mozne cist ze souboru " + path + "." + iex.Message);
                } catch (InvalidCastException ice) {
                    Prostredi.ShowMessage("Chyba v datech. " + ice.Message);
                }
                finally
                {
                    sr.Close();
                }
                
            }
        }

        /// <summary>
        /// Metoda nacita mereni z binarniho souboru. Pro kazde mereni jde o trojci hodnot daroveho typu Double.
        /// </summary>
        public void nacistZBinarnihoSouboru(String path)
        {
            Console.WriteLine("Binarni");
            kolekceMereni.Clear();
            FileInfo f = new FileInfo(path);
            long pocetzaznamu;
            try
            {
                pocetzaznamu = f.Length / 24; //Pocet zaznamu se urcije na zaklade velikosti souboru, ktera by mela byt 24 * pocet zaznamu bajtu
            }
            catch (InvalidCastException ice)
            {
                //Pokud pri deleni vznikne desetine cislo, soubor pravdepodobne neobsahuje mereni
                Prostredi.ShowMessage("Soubor " + path + " ma neocekavanou delku. Nemohu nacist data. " + ice.Message);
                return;
            }
            try
            {
                FileStream fs = new FileStream(path, FileMode.Open);
                BinaryReader br = new BinaryReader(fs);
                kolekceMereni.Clear();
                for (int i = 0; i < pocetzaznamu; i++)
                {
                    Mereni m = new Mereni();
                    //Trikrat se precte hodnota Double pro jedno mereni
                    m.X = br.ReadDouble();
                    m.Y = br.ReadDouble();
                    m.H = br.ReadDouble();
                    kolekceMereni.Add(m);
                }
                br.Close();
            }
            catch (IOException iex)
            {
                Prostredi.ShowMessage("Neni mozne cist ze souboru " + path + "." + iex.Message);
            }
        }

        /// <summary>
        /// Metoda nacita mereni z databaze. 
        /// </summary>
        public void nacistZDatabaze(String path) {
            Console.WriteLine("Databaze");
            kolekceMereni.Clear();
            try
            {
                //TODO: Vytvareny connectionString muze byt chybny. Vhodne pridat napovedu pro uzivatele nebo nabidnout vyber souboru z dialogoveho okna.
                string connectionString = "Provider=Microsoft.ACE.OLEDB.12.0;Data Source=" + path;
                OleDbConnection conn = new OleDbConnection(connectionString);
                conn.Open();
                OleDbCommand cmdread = new OleDbCommand();
                cmdread.Connection = conn;
                //V databazi musi but tabulka Mereni s presne dnaou strukturou. Pokud neni dojde k chybe.
                cmdread.CommandText = "SELECT * FROM Mereni";
                //Ziskame kurzor do ziskane tabulky
                OleDbDataReader reader = cmdread.ExecuteReader();
                //Kurzor se posunuje po radcich. Prvni pouziti metody Read presune kurzor na prvni radek v tabulce.
                while (reader.Read())
                {
                    Mereni m = new Mereni();
                    m.X = reader.GetDouble(1); //Cte se z druheho sloupce. Sloupec 1 obsahuje ID zaznamu.
                    m.Y = reader.GetDouble(2);
                    m.H = reader.GetDouble(3);
                    kolekceMereni.Add(m);
                }
                conn.Close();
            }
            catch (OleDbException ex)
            {
                Prostredi.ShowMessage(ex.Message);
            }
        }


        /// <summary>
        /// Metoda seradi mereni od nejblizsiho po nejvzdalenejsi od urcenych souradnic.
        /// </summary>
        private SortedList seraditMereni(double hledanybodX, double hledanybodY)
        {
            SortedList sl = new SortedList(); //Sorted list sam zaridi razeni dle urcene vzdalenosti
            foreach (Mereni m in kolekceMereni)
            {
                double deltaX = hledanybodX - m.X; //rozdil v souradnicich X
                double deltaY = hledanybodY - m.Y; //rozdil v souradnicich Y
                double vzdalenost = Math.Sqrt(Math.Pow(deltaX, 2) + Math.Pow(deltaY, 2)); //pouziti pythagorovy vety c^2 = a^2 + b^2
                sl.Add(vzdalenost, m); //Prvni parametr se pouziva pro razeni v kolekci
            }
            return sl;
        }

        /// <summary>
        /// Metoda spocita interpolaci.
        /// </summary>
        /// <param name="hledanybodX">Souradnice X pro misto interpolace.</param>
        /// <param name="hledanybodY">Souradnice Y pro misto interpolace.</param>
        /// <param name="exponent">Exponent pro umocneni vzdalenosti.</param>
        /// <param name="okoli">Pocet nejblizsich mereni, ktera se maji pouzit.</param>
        public Vysledek interpolovat(double hledanybodX, double hledanybodY, double exponent, int okoli)
        {
            double sumahodnotavaha = 0; //citatel - suma vah
            double sumavah = 0; //jmenovatel - suma vaha * hodnota
            SortedList sl = seraditMereni(hledanybodX, hledanybodY);
            //Pokud je pocet bodu v kolekci mensi nez pocet bodu, ktery se ma pouzit, pouziji se vsechna mereni z kolekce.
            int pocet = sl.Count;
            if (okoli < pocet) pocet = okoli;
            for (int i = 0; i < pocet; i++)
            {
                Mereni m = (Mereni) sl.GetByIndex(i); //Mereni je v SortedList urcene svym indexem
                double vzdalenost = (double) sl.GetKey(i); //Vzdalenost je urcena klicem - pouziva se pro razeni 
                double vaha = 1 / Math.Pow(vzdalenost, exponent);
                //Pricteni hodnoty do sum v citateli a jmenovateli
                sumahodnotavaha = sumahodnotavaha + (m.H * vaha);
                sumavah = sumavah + vaha;
            }
            //Trida Vysledek je pomocna a pouziva se pro lepsi srozumitelnost. Vhodnejsi je vyuzit parametry s modifikatory ref nebo out.
            Vysledek v = new Vysledek();
            if (sumavah > 0)
            {
                //Pokud nedojde k deleni nulou je pravdepododbne dobre spocitano.
                v.status = "OK";
                v.hodnota = sumahodnotavaha / sumavah;
                return v;
            }
            else {
                v.status = "KO";
                return v;
            }
        }
    }

    /// <summary>
    /// Trida je pouzita pro zaznam spocitane hodnoty a zarovne pro informaci zda nedoslo pri vypoctu k chybe.
    /// </summary>
    class Vysledek {
        public String status;
        public double hodnota;
    }

    /// <summary>
    /// Trida slouzi k nastaveni typou uzivatelskeho prostredi. Na zaklade typu je pak zprava pro uzivatele zobrazene bud v konzoli nebo v message boxu.
    /// </summary>
    class Prostredi {
        public static string ui = "Console";
        public static void ShowMessage(String message) { 
            if (ui.Equals("Console")) {
                Console.WriteLine(message);
            } else {
                MessageBox.Show(message);
            }
        }

    }
}
